==> spring-boot-ext-1.1.93_liqing-sources.jar: cn/com/duiba/boot/ext/autoconfigure/logger/LoggerAutoConfiguration.java <==
package cn.com.duiba.boot.ext.autoconfigure.logger;

import ch.qos.logback.classic.AsyncAppender;
import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.core.Appender;
import ch.qos.logback.core.FileAppender;
import ch.qos.logback.core.encoder.Encoder;
import ch.qos.logback.core.pattern.PatternLayoutEncoderBase;
import cn.com.duiba.boot.ext.autoconfigure.logger.logback.PatternLayoutEncoder;
import cn.com.duiba.boot.ext.autoconfigure.logger.logback.TimingLayoutEncoder;
import cn.com.duiba.boot.ext.autoconfigure.perftest.PerfTestRoutingDataSource;
import org.apache.commons.dbcp2.BasicDataSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.boot.autoconfigure.condition.ConditionalOnClass;
import org.springframework.boot.autoconfigure.condition.ConditionalOnResource;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationListener;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.event.ContextRefreshedEvent;
import org.springframework.core.io.ClassPathResource;

import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.joran.JoranConfigurator;

import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.util.*;

/**
 * Created by xugf on 2016/11/25
 */
@Configuration
@ConditionalOnClass({LoggerContext.class})
public class LoggerAutoConfiguration {
	private static final Logger logger = LoggerFactory.getLogger(LoggerAutoConfiguration.class);

	/**
	 * 附加logback配置文件(for loginsight日志监控)
	 * @return
	 */
	@Bean
	@ConditionalOnResource(resources="/logback/loginsight-logback.xml")
	public ApplicationListener loginsightLoggerConfiguar() {
		return new ApplicationListener<ContextRefreshedEvent>(){

			private boolean flag = true;

			@Override
			public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
				if(flag){
					try {
						LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
						JoranConfigurator configurator = new JoranConfigurator();
						configurator.setContext(loggerContext);
						//logback 添加新的配置文件
						configurator.doConfigure(new ClassPathResource("/logback/loginsight-logback.xml").getInputStream());

						Appender loginsightAppender = loggerContext.getLogger("loginsight_log").getAppender("ASYNC_LOGINSIGHT_APPENDER");
						loggerContext.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME).addAppender(loginsightAppender);
					} catch (Exception e) {
						logger.error(e.getMessage(), e);
					}
					flag = false;
				}
			}
		};
	}

	/**
	 * 数据库连接池检查报警,当连接池数目过满时打印报警日志
	 */
	@Configuration
	//@ConditionalOnResource(resources="/logback/loginsight-logback.xml")
	@ConditionalOnClass({BasicDataSource.class, LoggerContext.class})
	public static class dataSourceMonitorConfigurer{

		@Bean(name = "_dataSourceMonitor")
		public DataSourceMonitor _dataSourceMonitor(){
			return new DataSourceMonitor();
		}

		@Bean
		public ApplicationListener dataSourceMonitor() {
			return new ApplicationListener<ContextRefreshedEvent>() {

				@Override
				public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
					ApplicationContext context = applicationStartedEvent.getApplicationContext();
					DataSourceMonitor dataSourceMonitor = context.getBean("_dataSourceMonitor", DataSourceMonitor.class);
					dataSourceMonitor.startMonitorThread(context);
				}
			};
		}

		private static class DataSourceMonitor implements DisposableBean{
			private boolean initted = false;
			private Map<String,BasicDataSource> dataSourceMap;
			private Map<String,PerfTestRoutingDataSource> routingdataSourceMap;

			private Thread dataSourceMonitorThread;

			public synchronized void startMonitorThread(ApplicationContext context){
				if (!initted) {
					initted = true;
					dataSourceMap = context.getBeansOfType(BasicDataSource.class);
					routingdataSourceMap = context.getBeansOfType(PerfTestRoutingDataSource.class);

					if(dataSourceMap.isEmpty() && routingdataSourceMap.isEmpty()){
						return;
					}

					dataSourceMonitorThread = new Thread(){

						@Override
						public void run(){
							while(true){
								for(Map.Entry<String, BasicDataSource> entry : dataSourceMap.entrySet()){
									scanDataSource(entry.getKey(), entry.getValue());
								}
								for(Map.Entry<String, PerfTestRoutingDataSource> entry : routingdataSourceMap.entrySet()){
									scanDataSource(entry.getKey()+"_original", entry.getValue().getOriginalDataSource());
									scanDataSource(entry.getKey()+"_shade", entry.getValue().getShadeDataSource());
								}

								try {
									Thread.sleep(10000);//每隔10S检查一次,看连接数是否正常
								} catch (InterruptedException e) {
									break;
								}
								if(Thread.currentThread().isInterrupted()){
									break;
								}
							}
						}

					};

					dataSourceMonitorThread.start();
				}
			}

			private void scanDataSource(String key, BasicDataSource ds){
				if(ds != null && !ds.isClosed()) {
					int numIdleConnections = ds.getNumIdle();//池中有多少个空闲连接，它们可以被checkout
					int numBusyConnections = ds.getNumActive();//池中被checkout的连接有多少个
					int numConnections = numBusyConnections + numIdleConnections;////当前共几个连接
					int max = ds.getMaxTotal();
					if (numBusyConnections >= 3 && numConnections >= max - 3 && numBusyConnections >= numConnections - 3) {
						logger.error("datasource [" + key + "]'s connectionPool is too full,max:" + max + ",busy:" + numBusyConnections + ",idle:" + numIdleConnections);
					}
				}
			}

			@Override
			public synchronized void destroy() throws Exception {
				if(dataSourceMonitorThread != null) {
					dataSourceMonitorThread.interrupt();
				}
			}
		}

	}

	/**
	 * 自动把替换Encoder
	 * @return
	 */
	@Bean
	public ApplicationListener ReplaceEncoderConfiguar() {

		return new ApplicationListener<ContextRefreshedEvent>(){
			private boolean flag = true;
			//这个map用于缓存日志同步appender到日志异步appender
			@Override
			public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
				LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
				List<ch.qos.logback.classic.Logger> loggers = loggerContext.getLoggerList();
				for(ch.qos.logback.classic.Logger logger : loggers) {
					Iterator<Appender<ILoggingEvent>> iter = logger.iteratorForAppenders();

					while(iter.hasNext()){
						Appender<ILoggingEvent> appender = iter.next();
						if(appender instanceof FileAppender){
							FileAppender fileAppender = (FileAppender)appender;
							Encoder encoder = fileAppender.getEncoder();
							OutputStream outputStream = null;
							if(encoder != null && encoder instanceof ch.qos.logback.classic.encoder.PatternLayoutEncoder){
								String pattern = ((ch.qos.logback.classic.encoder.PatternLayoutEncoder) encoder).getPattern();
								try {
									Class object = encoder.getClass().getSuperclass().getSuperclass().getSuperclass();
									Field field = object.getDeclaredField("outputStream");
									field.setAccessible(true);
									outputStream = (OutputStream)field.get(encoder);
								} catch (NoSuchFieldException e) {
									logger.error(e.getMessage(),e);
								} catch (IllegalAccessException e) {
									logger.error(e.getMessage(),e);
								}
								TimingLayoutEncoder encoder1 = new TimingLayoutEncoder();
								if(outputStream != null){
									try {
										encoder1.init(outputStream);
									} catch (IOException e) {
										logger.error(e.getMessage(),e);
									}
								}

								
								boolean isoutputPatternAsHeader = ((ch.qos.logback.classic.encoder.PatternLayoutEncoder) encoder).isOutputPatternAsHeader();
								encoder1.setPattern(pattern);
								encoder1.setOutputPatternAsHeader(isoutputPatternAsHeader);
								fileAppender.setEncoder(encoder1);
							}

						}
					}

				}
			}
		};
	}

	/**
	 * 自动把logback日志输出转化为异步输出(如果已经配置异步输出则忽略)
	 * @return
	 */
	@Bean
	public ApplicationListener asyncLoggerConfiguar() {
		return new ApplicationListener<ContextRefreshedEvent>(){

			private boolean flag = true;
			//这个map用于缓存日志同步appender到日志异步appender
			private Map<FileAppender, AsyncAppender> syncAppender2asyncAppenderCachedMap = new HashMap<>();

			@Override
			public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
				if(flag){
					try {
						LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();

						//ch.qos.logback.classic.Logger rootLog = loggerContext.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME);
						List<ch.qos.logback.classic.Logger> loggers = loggerContext.getLoggerList();
						for(ch.qos.logback.classic.Logger logger : loggers) {
							switchLogAppend2Async(logger);
						}

						//这个map已经没用了，赋值null以便jvm回收
						syncAppender2asyncAppenderCachedMap = null;
					} catch (Exception e) {
						logger.error(e.getMessage(), e);
					}
					flag = false;
				}
			}

			/**
			 * 把当前log的同步输出appender转化为异步日志appender
			 * @param log
			 */
			private void switchLogAppend2Async(ch.qos.logback.classic.Logger log){
				Iterator<Appender<ILoggingEvent>> iter = log.iteratorForAppenders();
				List<Appender<ILoggingEvent>> asyncAppenders2Add = new ArrayList<>();
				List<Appender<ILoggingEvent>> appenders2Remove = new ArrayList<>();
				while(iter.hasNext()){
					Appender<ILoggingEvent> appender = iter.next();
					if(appender instanceof FileAppender){//如果是同步的FileAppender，则转换为异步输出Appender
						AsyncAppender asyncAppender = getCachedAsyncAppender((FileAppender)appender);
						asyncAppenders2Add.add(asyncAppender);
						appenders2Remove.add(appender);
					}
				}

				for(Appender<ILoggingEvent> syncAppender : appenders2Remove) {
					log.detachAppender(syncAppender);
				}
				for(Appender<ILoggingEvent> asyncAppender : asyncAppenders2Add) {
					log.addAppender(asyncAppender);
				}
			}

			/**
			 * 获得缓存的异步输出appender，如果未缓存，则构造之
			 * @return
			 */
			private AsyncAppender getCachedAsyncAppender(FileAppender fileAppender){
				AsyncAppender asyncAppender = syncAppender2asyncAppenderCachedMap.get(fileAppender);
				if(asyncAppender == null){
					asyncAppender = new AsyncAppender();
					asyncAppender.setContext((LoggerContext)LoggerFactory.getILoggerFactory());
					//不丢失日志. 默认地,如果队列的80%已满,则会丢弃TRACT、DEBUG、INFO级别的日志
					asyncAppender.setDiscardingThreshold(0);
					//更改默认的队列的深度,该值会影响性能.默认值为256
					asyncAppender.setQueueSize(512);
					//名字加入前缀

					//性能优化，只在必要的时候设置includeCallerData（经测试，输出行数等信息会增加10倍的rt）
					boolean includeCallerData = false;
					Encoder encoder = fileAppender.getEncoder();
					if(encoder != null && encoder instanceof PatternLayoutEncoderBase){
						PatternLayoutEncoderBase pencoder = (PatternLayoutEncoderBase)encoder;
						String pattern = pencoder.getPattern();
						if(pattern.contains("%L") || pattern.contains("%line")
								|| pattern.contains("%C") || pattern.contains("%class")
								|| pattern.contains("%M") || pattern.contains("%method")
								|| pattern.contains("%F") || pattern.contains("%file")
								|| pattern.contains("%caller")){
							includeCallerData = true;
						}
					}

					//这里设为true才能正确打印出%line等信息,如果不需要打印%line，则这里不需要配为true
					asyncAppender.setIncludeCallerData(includeCallerData);

					asyncAppender.setName("DuibaBootAutoAsync_"+fileAppender.getName());
					asyncAppender.addAppender(fileAppender);
					//这里只管start，LoggerContext.stop()被调用时会自动调用这个asyncAppender的stop方法。
					asyncAppender.start();

					syncAppender2asyncAppenderCachedMap.put(fileAppender, asyncAppender);
				}

				return asyncAppender;
			}
		};
	}



}

==> spring-boot-ext-1.1.101_liqing-sources.jar: cn/com/duiba/boot/ext/autoconfigure/logger/LoggerAutoConfiguration.java <==
package cn.com.duiba.boot.ext.autoconfigure.logger;

import ch.qos.logback.classic.AsyncAppender;
import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.core.Appender;
import ch.qos.logback.core.Context;
import ch.qos.logback.core.FileAppender;
import ch.qos.logback.core.encoder.Encoder;
import ch.qos.logback.core.pattern.PatternLayoutEncoderBase;
import cn.com.duiba.boot.ext.autoconfigure.cat.CatSpringDataRedisPlugin;
import cn.com.duiba.boot.ext.autoconfigure.logger.logback.PatternLayoutEncoder;
import cn.com.duiba.boot.ext.autoconfigure.logger.logback.TimingLayoutEncoder;
import cn.com.duiba.boot.ext.autoconfigure.perftest.PerfTestRoutingDataSource;
import org.apache.commons.dbcp2.BasicDataSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.aop.MethodBeforeAdvice;
import org.springframework.aop.framework.ProxyFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.boot.autoconfigure.condition.ConditionalOnClass;
import org.springframework.boot.autoconfigure.condition.ConditionalOnResource;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationListener;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.event.ContextRefreshedEvent;
import org.springframework.core.io.ClassPathResource;

import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.joran.JoranConfigurator;
import org.springframework.data.redis.connection.jedis.JedisConnection;

import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.*;

/**
 * Created by xugf on 2016/11/25
 */
@Configuration
@ConditionalOnClass({LoggerContext.class})
public class LoggerAutoConfiguration {
	private static final Logger logger = LoggerFactory.getLogger(LoggerAutoConfiguration.class);

	/**
	 * 附加logback配置文件(for loginsight日志监控)
	 * @return
	 */
	@Bean
	@ConditionalOnResource(resources="/logback/loginsight-logback.xml")
	public ApplicationListener loginsightLoggerConfiguar() {
		return new ApplicationListener<ContextRefreshedEvent>(){

			private boolean flag = true;

			@Override
			public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
				if(flag){
					try {
						LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
						JoranConfigurator configurator = new JoranConfigurator();
						configurator.setContext(loggerContext);
						//logback 添加新的配置文件
						configurator.doConfigure(new ClassPathResource("/logback/loginsight-logback.xml").getInputStream());

						Appender loginsightAppender = loggerContext.getLogger("loginsight_log").getAppender("ASYNC_LOGINSIGHT_APPENDER");
						loggerContext.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME).addAppender(loginsightAppender);
					} catch (Exception e) {
						logger.error(e.getMessage(), e);
					}
					flag = false;
				}
			}
		};
	}

	/**
	 * 数据库连接池检查报警,当连接池数目过满时打印报警日志
	 */
	@Configuration
	//@ConditionalOnResource(resources="/logback/loginsight-logback.xml")
	@ConditionalOnClass({BasicDataSource.class, LoggerContext.class})
	public static class dataSourceMonitorConfigurer{

		@Bean(name = "_dataSourceMonitor")
		public DataSourceMonitor _dataSourceMonitor(){
			return new DataSourceMonitor();
		}

		@Bean
		public ApplicationListener dataSourceMonitor() {
			return new ApplicationListener<ContextRefreshedEvent>() {

				@Override
				public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
					ApplicationContext context = applicationStartedEvent.getApplicationContext();
					DataSourceMonitor dataSourceMonitor = context.getBean("_dataSourceMonitor", DataSourceMonitor.class);
					dataSourceMonitor.startMonitorThread(context);
				}
			};
		}

		private static class DataSourceMonitor implements DisposableBean{
			private boolean initted = false;
			private Map<String,BasicDataSource> dataSourceMap;
			private Map<String,PerfTestRoutingDataSource> routingdataSourceMap;

			private Thread dataSourceMonitorThread;

			public synchronized void startMonitorThread(ApplicationContext context){
				if (!initted) {
					initted = true;
					dataSourceMap = context.getBeansOfType(BasicDataSource.class);
					routingdataSourceMap = context.getBeansOfType(PerfTestRoutingDataSource.class);

					if(dataSourceMap.isEmpty() && routingdataSourceMap.isEmpty()){
						return;
					}

					dataSourceMonitorThread = new Thread(){

						@Override
						public void run(){
							while(true){
								for(Map.Entry<String, BasicDataSource> entry : dataSourceMap.entrySet()){
									scanDataSource(entry.getKey(), entry.getValue());
								}
								for(Map.Entry<String, PerfTestRoutingDataSource> entry : routingdataSourceMap.entrySet()){
									scanDataSource(entry.getKey()+"_original", entry.getValue().getOriginalDataSource());
									scanDataSource(entry.getKey()+"_shade", entry.getValue().getShadeDataSource());
								}

								try {
									Thread.sleep(10000);//每隔10S检查一次,看连接数是否正常
								} catch (InterruptedException e) {
									break;
								}
								if(Thread.currentThread().isInterrupted()){
									break;
								}
							}
						}

					};

					dataSourceMonitorThread.start();
				}
			}

			private void scanDataSource(String key, BasicDataSource ds){
				if(ds != null && !ds.isClosed()) {
					int numIdleConnections = ds.getNumIdle();//池中有多少个空闲连接，它们可以被checkout
					int numBusyConnections = ds.getNumActive();//池中被checkout的连接有多少个
					int numConnections = numBusyConnections + numIdleConnections;////当前共几个连接
					int max = ds.getMaxTotal();
					if (numBusyConnections >= 3 && numConnections >= max - 3 && numBusyConnections >= numConnections - 3) {
						logger.error("datasource [" + key + "]'s connectionPool is too full,max:" + max + ",busy:" + numBusyConnections + ",idle:" + numIdleConnections);
					}
				}
			}

			@Override
			public synchronized void destroy() throws Exception {
				if(dataSourceMonitorThread != null) {
					dataSourceMonitorThread.interrupt();
				}
			}
		}

	}

	/**
	 * 自动把替换Encoder
	 * @return
	 */
	@Bean
	public ApplicationListener ReplaceEncoderConfiguar() {

		return new ApplicationListener<ContextRefreshedEvent>(){
			private OutputStream outputStream = null;
			private Timer timer = new Timer();
			@Override
			public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
				LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();
				List<ch.qos.logback.classic.Logger> loggers = loggerContext.getLoggerList();
				for(ch.qos.logback.classic.Logger logger : loggers) {
					Iterator<Appender<ILoggingEvent>> iter = logger.iteratorForAppenders();

					while(iter.hasNext()){
						Appender<ILoggingEvent> appender = iter.next();
						if(appender instanceof FileAppender){
							FileAppender fileAppender = (FileAppender)appender;
							Encoder encoder = fileAppender.getEncoder();

							if(encoder != null && encoder instanceof ch.qos.logback.classic.encoder.PatternLayoutEncoder){
//								String pattern = ((ch.qos.logback.classic.encoder.PatternLayoutEncoder) encoder).getPattern();
								try {
									Class object = encoder.getClass().getSuperclass().getSuperclass().getSuperclass();
									Field field = object.getDeclaredField("outputStream");
									field.setAccessible(true);
									outputStream = (OutputStream)field.get(encoder);
								} catch (NoSuchFieldException e) {
									logger.error(e.getMessage(),e);
								} catch (IllegalAccessException e) {
									logger.error(e.getMessage(),e);
								}
								ProxyFactory proxyFactory=new ProxyFactory();
								proxyFactory.setTarget(fileAppender);

								proxyFactory.addAdvice(new MethodBeforeAdvice() {
									private static final int intevalPeriod = 60 * 1000;//second
									private static final int timeDelay = 60 * 1000;
									TimerTask timerTask = new TimerTask() {
										@Override
										public void run(){

											if(outputStream != null)
												try {
													outputStream.flush();
												} catch (IOException e) {
													throw new RuntimeException(e);
												}
										}
									};
									@Override
									public void before(Method method, Object[] args, Object target)
											throws Throwable {
										timer.schedule(timerTask,timeDelay,intevalPeriod);
									}
								});
								FileAppender newFileAppender = (FileAppender) proxyFactory.getProxy();//getClass().getClassLoader()
								newFileAppender.setEncoder(encoder);

//								TimingLayoutEncoder encoder1 = new TimingLayoutEncoder();
//								if(outputStream != null){
//									try {
//										encoder1.init(outputStream);
//									} catch (IOException e) {
//										logger.error(e.getMessage(),e);
//									}
//								}
//								Context context = encoder.getContext();
//								boolean isoutputPatternAsHeader = ((ch.qos.logback.classic.encoder.PatternLayoutEncoder) encoder).isOutputPatternAsHeader();
//								encoder1.setPattern(pattern);
//								encoder1.setContext(context);
//								encoder1.setImmediateFlush(((ch.qos.logback.classic.encoder.PatternLayoutEncoder) encoder).isImmediateFlush());
//								encoder1.setOutputPatternAsHeader(isoutputPatternAsHeader);
//								encoder1.start();
//								fileAppender.setEncoder(encoder1);
								//利用spring的API,创建接口代理工厂

//								if(lastJedisConnectionProxyClass != null && lastJedisConnectionProxyClass != newConnection.getClass()){
//									logger.error("JedisConnectionProxyClass is not same，this is spring's bug,please upgrade spring-boot's version to 1.3.8.RELEASE or higher! {},{},{},{}",new Object[]{getClass().getClassLoader(), newConnection.getClass().getClassLoader(),lastJedisConnectionProxyClass, newConnection.getClass()});//如果抱这个错，则有perm区内存溢出的风险，需要关注
//								}
//								lastJedisConnectionProxyClass = newConnection.getClass();
//
//								return newConnection;
							}

						}
					}

				}
			}
		};
	}

	/**
	 * 自动把logback日志输出转化为异步输出(如果已经配置异步输出则忽略)
	 * @return
	 */
	@Bean
	public ApplicationListener asyncLoggerConfiguar() {
		return new ApplicationListener<ContextRefreshedEvent>(){

			private boolean flag = true;
			//这个map用于缓存日志同步appender到日志异步appender
			private Map<FileAppender, AsyncAppender> syncAppender2asyncAppenderCachedMap = new HashMap<>();

			@Override
			public void onApplicationEvent(ContextRefreshedEvent applicationStartedEvent) {
				if(flag){
					try {
						LoggerContext loggerContext = (LoggerContext) LoggerFactory.getILoggerFactory();

						//ch.qos.logback.classic.Logger rootLog = loggerContext.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME);
						List<ch.qos.logback.classic.Logger> loggers = loggerContext.getLoggerList();
						for(ch.qos.logback.classic.Logger logger : loggers) {
							switchLogAppend2Async(logger);
						}

						//这个map已经没用了，赋值null以便jvm回收
						syncAppender2asyncAppenderCachedMap = null;
					} catch (Exception e) {
						logger.error(e.getMessage(), e);
					}
					flag = false;
				}
			}

			/**
			 * 把当前log的同步输出appender转化为异步日志appender
			 * @param log
			 */
			private void switchLogAppend2Async(ch.qos.logback.classic.Logger log){
				Iterator<Appender<ILoggingEvent>> iter = log.iteratorForAppenders();
				List<Appender<ILoggingEvent>> asyncAppenders2Add = new ArrayList<>();
				List<Appender<ILoggingEvent>> appenders2Remove = new ArrayList<>();
				while(iter.hasNext()){
					Appender<ILoggingEvent> appender = iter.next();
					if(appender instanceof FileAppender){//如果是同步的FileAppender，则转换为异步输出Appender
						AsyncAppender asyncAppender = getCachedAsyncAppender((FileAppender)appender);
						asyncAppenders2Add.add(asyncAppender);
						appenders2Remove.add(appender);
					}
				}

				for(Appender<ILoggingEvent> syncAppender : appenders2Remove) {
					log.detachAppender(syncAppender);
				}
				for(Appender<ILoggingEvent> asyncAppender : asyncAppenders2Add) {
					log.addAppender(asyncAppender);
				}
			}

			/**
			 * 获得缓存的异步输出appender，如果未缓存，则构造之
			 * @return
			 */
			private AsyncAppender getCachedAsyncAppender(FileAppender fileAppender){
				AsyncAppender asyncAppender = syncAppender2asyncAppenderCachedMap.get(fileAppender);
				if(asyncAppender == null){
					asyncAppender = new AsyncAppender();
					asyncAppender.setContext((LoggerContext)LoggerFactory.getILoggerFactory());
					//不丢失日志. 默认地,如果队列的80%已满,则会丢弃TRACT、DEBUG、INFO级别的日志
					asyncAppender.setDiscardingThreshold(0);
					//更改默认的队列的深度,该值会影响性能.默认值为256
					asyncAppender.setQueueSize(512);
					//名字加入前缀

					//性能优化，只在必要的时候设置includeCallerData（经测试，输出行数等信息会增加10倍的rt）
					boolean includeCallerData = false;
					Encoder encoder = fileAppender.getEncoder();
					if(encoder != null && encoder instanceof PatternLayoutEncoderBase){
						PatternLayoutEncoderBase pencoder = (PatternLayoutEncoderBase)encoder;
						String pattern = pencoder.getPattern();
						if(pattern.contains("%L") || pattern.contains("%line")
								|| pattern.contains("%C") || pattern.contains("%class")
								|| pattern.contains("%M") || pattern.contains("%method")
								|| pattern.contains("%F") || pattern.contains("%file")
								|| pattern.contains("%caller")){
							includeCallerData = true;
						}
					}

					//这里设为true才能正确打印出%line等信息,如果不需要打印%line，则这里不需要配为true
					asyncAppender.setIncludeCallerData(includeCallerData);

					asyncAppender.setName("DuibaBootAutoAsync_"+fileAppender.getName());
					asyncAppender.addAppender(fileAppender);
					//这里只管start，LoggerContext.stop()被调用时会自动调用这个asyncAppender的stop方法。
					asyncAppender.start();

					syncAppender2asyncAppenderCachedMap.put(fileAppender, asyncAppender);
				}

				return asyncAppender;
			}
		};
	}



}
